﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;

using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.API.Config;
using Vintagestory.API.Datastructures;
using Vintagestory.API.MathTools;
using Vintagestory.Common;


namespace Automap
{
	public class JsonGenerator
	{
		private ICoreClientAPI ClientAPI { get; set; }
		private ILogger Logger { get; set; }
		private string path { get; set; }
		private readonly int chunkSize;

		internal JArray ColumnMeta_fieldNames,PointsOfInterest_fieldNames,EntitiesOfInterest_fieldNames;

		public JsonGenerator(ICoreClientAPI _ClientAPI,  ILogger _Logger, string _path )
		{
		this.ClientAPI = _ClientAPI;
		this.Logger = _Logger;
		this.path = _path;
		this.chunkSize = ClientAPI.World.BlockAccessor.ChunkSize;

		ColumnMeta_fieldNames = Dynamic_Names<ColumnMeta>( );
		PointsOfInterest_fieldNames = Dynamic_Names<PointOfInterest>( );
		EntitiesOfInterest_fieldNames = Dynamic_Names<EntityOfInterest>( );
		Logger.VerboseDebug("JSON Ready");
		}


		/// <summary>
		/// Generates the JSON Metadata. (in Map object format )
		/// </summary>
		public void GenerateJSONMetadata(ColumnsMetadata chunkTopMetadata, Vec2i startChunkColumn, PointsOfInterest POIs, EntitiesOfInterest EOIs, Dictionary<int, string> RockIdCodes )
		{
		string jsonFilename = Path.Combine(path, "Metadata.js");

		StreamWriter stream = new StreamWriter(jsonFilename, false, Encoding.UTF8);

		using (stream) {
		JsonTextWriter jsonWriter = new JsonTextWriter(stream);

		jsonWriter.Formatting = Formatting.None;
		jsonWriter.StringEscapeHandling = StringEscapeHandling.EscapeHtml;
		jsonWriter.Indentation = 0;
		//jsonWriter.AutoCompleteOnClose = true;
		jsonWriter.QuoteChar = '\'';
		jsonWriter.DateFormatHandling = DateFormatHandling.IsoDateFormat;
		jsonWriter.DateTimeZoneHandling = DateTimeZoneHandling.Utc;

				/*
ViewFrame.chunks
vfc ={};
// the basic metadata right now
vfc.edges = [northmostedge, eastmost, west, south];
vfc.chunkMetadataNames=[...]; // the implementation with reflection
vfc.chunkMetadata = new Map([['the keys we have', [...metadatas in the same order as the names, stringified(maybe not for rocks thats complicated)],...]);
*/

		using (jsonWriter) {
		jsonWriter.WriteRaw("ViewFrame.chunks={};\n");
		jsonWriter.WriteRaw("ViewFrame.chunks.worldSeedNum=");
		jsonWriter.WriteValue(ClientAPI.World.Seed);
		jsonWriter.WriteRaw(";\n");

		jsonWriter.WriteRaw("ViewFrame.chunks.genTime=");
		jsonWriter.WriteValue(DateTimeOffset.UtcNow);
		jsonWriter.WriteRaw(";\n");

		jsonWriter.WriteRaw("ViewFrame.chunks.startCoords=");
		jsonWriter.WriteStartArray( );
		jsonWriter.WriteValue(startChunkColumn.X);
		jsonWriter.WriteValue(startChunkColumn.Y);
		jsonWriter.WriteEndArray( );
		jsonWriter.WriteRaw(";\n");

		jsonWriter.WriteRaw("ViewFrame.chunks.chunkSize=");
		jsonWriter.WriteValue(chunkSize);
		jsonWriter.WriteRaw(";\n");

		jsonWriter.WriteRaw("ViewFrame.chunks.edges=");//whats NEWS
		jsonWriter.WriteStartArray( );
		jsonWriter.WriteValue(chunkTopMetadata.North_mostChunk);
		jsonWriter.WriteValue(chunkTopMetadata.East_mostChunk);
		jsonWriter.WriteValue(chunkTopMetadata.South_mostChunk);		
		jsonWriter.WriteValue(chunkTopMetadata.West_mostChunk);		
		jsonWriter.WriteEndArray( );
		jsonWriter.WriteRaw(";\n");
							
		jsonWriter.WriteRaw("ViewFrame.chunks.chunkMetadataNames=");		
		ColumnMeta_fieldNames.WriteTo(jsonWriter);
		jsonWriter.WriteRaw(";\n");

		jsonWriter.WriteRaw("ViewFrame.chunks.pointsOfInterestNames=");		
		PointsOfInterest_fieldNames.WriteTo(jsonWriter);
		jsonWriter.WriteRaw(";\n");

		jsonWriter.WriteRaw("ViewFrame.chunks.entityOfInterestNames=");		
		EntitiesOfInterest_fieldNames.WriteTo(jsonWriter);		
		jsonWriter.WriteRaw(";\n");

		//MAP object format - [key, value]: key is "x_y"
		jsonWriter.WriteRaw("ViewFrame.chunks.chunkMetadata=");
		jsonWriter.WriteStartConstructor("Map");
		jsonWriter.WriteStartArray( );//An array of... 2-component arrays


		foreach (var shard in chunkTopMetadata) {
		//TODO:Flatten to array of values...Dynamism!
		jsonWriter.WriteStartArray( );//Start tuple
		jsonWriter.WriteValue($"{shard.Location.X}_{shard.Location.Y}");//Key of Tuple

		jsonWriter.WriteStartObject( );
		jsonWriter.WritePropertyName("prettyCoord");
		jsonWriter.WriteValue(shard.Location.PrettyCoords(ClientAPI));

		jsonWriter.WritePropertyName("chunkAge");
		jsonWriter.WriteValue(shard.ChunkAge);

		jsonWriter.WritePropertyName("temp");
		jsonWriter.WriteValue(shard.Temperature);

		jsonWriter.WritePropertyName("YMax");
		jsonWriter.WriteValue(shard.YMax);

		jsonWriter.WritePropertyName("fert");
		jsonWriter.WriteValue(shard.Fertility);

		jsonWriter.WritePropertyName("forestDens");
		jsonWriter.WriteValue(shard.ForestDensity);

		jsonWriter.WritePropertyName("rain");
		jsonWriter.WriteValue(shard.Rainfall);

		jsonWriter.WritePropertyName("shrubDens");
		jsonWriter.WriteValue(shard.ShrubDensity);

		jsonWriter.WritePropertyName("airBlocks");
		jsonWriter.WriteValue(shard.AirBlocks);

		jsonWriter.WritePropertyName("nonAirBlocks");
		jsonWriter.WriteValue(shard.NonAirBlocks);

		//TODO: Heightmap ?
		//Start rockMap ; FOR a Ratio....on tooltip GUI
		jsonWriter.WritePropertyName("rockRatio");
		jsonWriter.WriteStartConstructor("Map");
		jsonWriter.WriteStartArray( );
		foreach (var rockEntry in shard.RockRatio) {
		var rockBlock = ClientAPI.World.GetBlock(rockEntry.Key);
		jsonWriter.WriteStartArray( );
		jsonWriter.WriteValue(rockBlock.Code.Path);
		jsonWriter.WriteValue(rockEntry.Value);//Total per chunk-column
		jsonWriter.WriteEndArray( );
		}
		jsonWriter.WriteEndArray( );
		jsonWriter.WriteEndConstructor( );//end rock-map

		jsonWriter.WriteEndObject( );//end Map value: {Object}
		jsonWriter.WriteEndArray( );//end Tuple
		}

		jsonWriter.WriteEndArray( );//Enclose tuples of chunkMetadata
		jsonWriter.WriteEndConstructor( );//Close constructor of Map (chunkMetadata)
		jsonWriter.WriteRaw(";\n");

		jsonWriter.WriteRaw("ViewFrame.chunks.pointsOfInterest=");
		jsonWriter.WriteStartConstructor("Map");
		jsonWriter.WriteStartArray( );//An array of... 2-component arrays

		foreach (var poi in POIs) {
		EmitJsonMap(poi, jsonWriter);
		}

		foreach (var poi in EOIs) {
		EmitJsonMap(poi, jsonWriter);
		}

		jsonWriter.WriteEndArray( );
		jsonWriter.WriteEndConstructor( );
		jsonWriter.WriteRaw(";\n");

		jsonWriter.WriteWhitespace("\n");
		jsonWriter.WriteComment("============= BlockID's for Rockmap / Rock-ratios ===============");
		jsonWriter.WriteWhitespace("\n");

		jsonWriter.WriteRaw("ViewFrame.chunks.rock_Lookup =");
		jsonWriter.WriteStartConstructor("Map");
		jsonWriter.WriteStartArray( );//An array of... 2-component arrays

		foreach (var entry in RockIdCodes) {
		var block = ClientAPI.World.GetBlock(entry.Key);
		//	stream.Write("{0},", Lang.GetUnformatted(block.Code.Path));
		jsonWriter.WriteStartArray( );
		jsonWriter.WriteValue(block.Code.Path);

		jsonWriter.WriteStartObject( );
		jsonWriter.WritePropertyName("assetCode");
		jsonWriter.WriteValue(entry.Value);

		jsonWriter.WritePropertyName("name");
		jsonWriter.WriteValue(Lang.GetUnformatted(block.Code.Path));
		//}

		jsonWriter.WriteEndObject( );
		jsonWriter.WriteEndArray( );
		}
		jsonWriter.WriteEndArray( );
		jsonWriter.WriteEndConstructor( );
		//stream.Write("]);\n");
		jsonWriter.WriteRaw(";\n");

		jsonWriter.Flush( );
		}
		}

		}

		public void EmitJsonMap(PointOfInterest @this, JsonTextWriter jsonWriter)
		{
		jsonWriter.WriteStartArray( );
		jsonWriter.WriteValue($"{@this.Location.X}_{@this.Location.Z}");

		jsonWriter.WriteStartArray( );
		//jsonWriter.WriteValue(@this.Name);

		jsonWriter.WriteValue(@this.Location.X);
		jsonWriter.WriteValue(@this.Location.Z); //Y is HEIGHT from Mantle... Z is Right from edge of world

		jsonWriter.WriteValue(@this.Location.PrettyCoords(ClientAPI));

		jsonWriter.WriteValue(@this.Notes.Replace('\\', ' '));//put more escaping in Java-script if needed

		jsonWriter.WriteValue(@this.Timestamp);

		jsonWriter.WriteEndArray( );
		jsonWriter.WriteEndArray( );
		}

		public void EmitJsonMap(EntityOfInterest @this, JsonTextWriter jsonWriter)
		{
		jsonWriter.WriteStartArray( );
		jsonWriter.WriteValue($"{@this.Location.X}_{@this.Location.Z}");


		jsonWriter.WriteStartArray( );
		//jsonWriter.WriteValue(@this.Name);

		jsonWriter.WriteValue(@this.Location.X);
		jsonWriter.WriteValue(@this.Location.Z); //Y is HEIGHT from Mantle... Z is Right from edge of world

		jsonWriter.WriteValue(@this.Location.PrettyCoords(ClientAPI));

		jsonWriter.WriteValue(@this.Notes.Replace('\\', ' '));//put more escaping in Java-script if needed

		jsonWriter.WriteValue(@this.Timestamp);

		//jsonWriter.WriteValue(@this.EntityId);		

		jsonWriter.WriteEndArray( );
		jsonWriter.WriteEndArray( );
		}


		internal JArray Dynamic_Names<TData>( ) where TData : struct
		{
		Dictionary<byte, string> fieldNames = new Dictionary<byte, string>( );

		foreach (var fieldInfo in typeof(TData).GetFields(BindingFlags.Instance | BindingFlags.Public)) {		
		DisplayNameAttribute displayName = fieldInfo.GetCustomAttribute<DisplayNameAttribute>();
			if (displayName != null) 
			{
				if (!fieldNames.ContainsKey(displayName.order)) 
				{//No duplicates, no overwrites
				fieldNames.Add(displayName.order, displayName.name);
				}				
			}
		}

		return new JArray(fieldNames.OrderBy(kf => kf.Key).Select(kf => kf.Value).ToArray());	
		}


	}
}

